@@ -136,6 +136,9 @@ static int crtools_prepare_shared(void)
 	if (collect_inet_sockets())
 		return -1;
 
+	if (collect_binfmt_misc())
+		return -1;
+
 	if (tty_prep_fds())
 		return -1;
 
@@ -704,7 +707,7 @@ static int restore_one_zombie(CoreEntry *core)
 	prctl(PR_SET_NAME, (long)(void *)core->tc->comm, 0, 0, 0);
 
 	if (task_entries != NULL) {
-		restore_finish_stage(CR_STATE_RESTORE);
+		restore_finish_stage(task_entries, CR_STATE_RESTORE);
 		zombie_prepare_signals();
 	}
 
@@ -786,7 +789,7 @@ static int restore_one_task(int pid, CoreEntry *core)
 			return -1;
 		}
 
-		restore_finish_stage(CR_STATE_RESTORE);
+		restore_finish_stage(task_entries, CR_STATE_RESTORE);
 		if (wait_on_helpers_zombies()) {
 			pr_err("failed to wait on helpers and zombies\n");
 			ret = -1;
@@ -1278,7 +1281,7 @@ static int restore_task_with_children(void *_arg)
 
 	/* Wait prepare_userns */
 	if (current->parent == NULL &&
-            restore_finish_stage(CR_STATE_RESTORE_NS) < 0)
+            restore_finish_stage(task_entries, CR_STATE_RESTORE_NS) < 0)
 			goto err;
 
 	/*
@@ -1314,7 +1317,7 @@ static int restore_task_with_children(void *_arg)
 		if (root_prepare_shared())
 			goto err;
 
-		if (restore_finish_stage(CR_STATE_RESTORE_SHARED) < 0)
+		if (restore_finish_stage(task_entries, CR_STATE_RESTORE_SHARED) < 0)
 			goto err;
 	}
 
@@ -1351,7 +1354,7 @@ static int restore_task_with_children(void *_arg)
 		fini_restore_mntns();
 	}
 
-	if (restore_finish_stage(CR_STATE_FORKING) < 0)
+	if (restore_finish_stage(task_entries, CR_STATE_FORKING) < 0)
 		goto err;
 
 	if (restore_one_task(current->pid.virt, ca->core))
@@ -1533,7 +1536,7 @@ static void finalize_restore(void)
 			continue;
 
 		/* Unmap the restorer blob */
-		ctl = parasite_prep_ctl(pid, NULL);
+		ctl = parasite_prep_ctl(pid, 0);
 		if (ctl == NULL)
 			continue;
 
@@ -1541,7 +1544,8 @@ static void finalize_restore(void)
 
 		xfree(ctl);
 
-		if (item->pid.state == TASK_STOPPED)
+		if ((item->pid.state == TASK_STOPPED) ||
+				(opts.final_state == TASK_STOPPED))
 			kill(item->pid.real, SIGSTOP);
 	}
 }
@@ -2053,6 +2057,9 @@ static int prepare_itimers_from_fd(int pid, struct task_restore_args *args)
 	struct cr_img *img;
 	ItimerEntry *ie;
 
+	if (!deprecated_ok("Itimers"))
+		return -1;
+
 	img = open_image(CR_FD_ITIMERS, O_RSTR, pid);
 	if (!img)
 		return -1;
@@ -2175,6 +2182,9 @@ static int prepare_posix_timers_from_fd(int pid, struct task_restore_args *ta)
 	int ret = -1;
 	struct restore_posix_timer *t;
 
+	if (!deprecated_ok("Posix timers"))
+		return -1;
+
 	img = open_image(CR_FD_POSIX_TIMERS, O_RSTR, pid);
 	if (!img)
 		return -1;
@@ -2361,6 +2371,9 @@ static int prepare_rlimits_from_fd(int pid, struct task_restore_args *ta)
 	int ret;
 	struct cr_img *img;
 
+	if (!deprecated_ok("Rlimits"))
+		return -1;
+
 	/*
 	 * Old image -- read from the file.
 	 */
@@ -2741,6 +2754,8 @@ static int sigreturn_restore(pid_t pid, struct task_restore_args *task_args, uns
 	unsigned long creds_pos = 0;
 	unsigned long creds_pos_next;
 
+	sigset_t blockmask;
+
 	pr_info("Restore via sigreturn\n");
 
 	/* pr_info_vma_list(&self_vma_list); */
@@ -2868,6 +2883,15 @@ static int sigreturn_restore(pid_t pid, struct task_restore_args *task_args, uns
 	}
 
 	task_args->breakpoint = &rsti(current)->breakpoint;
+
+	sigemptyset(&blockmask);
+	sigaddset(&blockmask, SIGCHLD);
+
+	if (sigprocmask(SIG_BLOCK, &blockmask, NULL) == -1) {
+		pr_perror("Can not set mask of blocked signals");
+		return -1;
+	}
+
 	task_args->task_entries = rst_mem_remap_ptr(task_entries_pos, RM_SHREMAP);
 
 	task_args->premmapped_addr = (unsigned long)rsti(current)->premmapped_addr;
@@ -2910,6 +2934,7 @@ static int sigreturn_restore(pid_t pid, struct task_restore_args *task_args, uns
 	for (i = 0; i < current->nr_threads; i++) {
 		CoreEntry *tcore;
 		struct rt_sigframe *sigframe;
+		k_rtsigset_t *blkset = NULL;
 
 		thread_args[i].pid = current->threads[i].virt;
 		thread_args[i].siginfo_n = siginfo_priv_nr[i];
@@ -2921,8 +2946,12 @@ static int sigreturn_restore(pid_t pid, struct task_restore_args *task_args, uns
 		if (thread_args[i].pid == pid) {
 			task_args->t = thread_args + i;
 			tcore = core;
-		} else
+			blkset = (void *)&tcore->tc->blk_sigset;
+		} else {
 			tcore = current->core[i];
+			if (tcore->thread_core->has_blk_sigset)
+				blkset = (void *)&tcore->thread_core->blk_sigset;
+		}
 
 		if ((tcore->tc || tcore->ids) && thread_args[i].pid != pid) {
 			pr_err("Thread has optional fields present %d\n",
@@ -2959,7 +2988,7 @@ static int sigreturn_restore(pid_t pid, struct task_restore_args *task_args, uns
 		thread_args[i].mz = mz + i;
 		sigframe = (struct rt_sigframe *)&mz[i].rt_sigframe;
 
-		if (construct_sigframe(sigframe, sigframe, tcore))
+		if (construct_sigframe(sigframe, sigframe, blkset, tcore))
 			goto err;
 
 		if (thread_args[i].pid != pid)
