@@ -78,7 +78,9 @@ struct cr_fd_desc_tmpl imgset_template[CR_FD_MAX] = {
 	FD_ENTRY_F(IP6TABLES,	"ip6tables-%d", O_NOBUF),
 	FD_ENTRY_F(TMPFS_IMG,	"tmpfs-%d.tar.gz", O_NOBUF),
 	FD_ENTRY_F(TMPFS_DEV,	"tmpfs-dev-%d.tar.gz", O_NOBUF),
-	FD_ENTRY(BINFMT_MISC,	"binfmt-misc-%d"),
+	FD_ENTRY_F(AUTOFS,	"autofs-%d", O_NOBUF),
+	FD_ENTRY(BINFMT_MISC_OLD, "binfmt-misc-%d"),
+	FD_ENTRY(BINFMT_MISC,	"binfmt-misc"),
 	FD_ENTRY(TTY_FILES,	"tty"),
 	FD_ENTRY(TTY_INFO,	"tty-info"),
 	FD_ENTRY_F(TTY_DATA,	"tty-data", O_NOBUF),
