@@ -47,11 +47,6 @@ struct irmap_path_opt {
 	struct irmap *ir;
 };
 
-struct external {
-	struct list_head node;
-	char *id;
-};
-
 struct cr_options {
 	int			final_state;
 	char			*show_dump_file;
@@ -103,6 +98,7 @@ struct cr_options {
 	bool			enable_external_masters;
 	bool			aufs;		/* auto-deteced, not via cli */
 	bool			overlayfs;
+	bool			has_binfmt_misc; /* auto-detected */
 	size_t			ghost_limit;
 	struct list_head	irmap_scan_paths;
 	bool			lsm_supplied;
@@ -111,12 +107,18 @@ struct cr_options {
 	unsigned int		empty_ns;
 	bool			tcp_skip_in_flight;
 	char			*work_dir;
+
+	/*
+	 * When we scheduler for removal some functionality we first
+	 * deprecate it and it sits in criu for some time. By default
+	 * the deprecated stuff is not working, but it's still possible
+	 * to turn one ON while the code is in.
+	 */
+	bool			deprecated_ok;
 };
 
 extern struct cr_options opts;
 
 extern void init_opts(void);
 
-extern int add_external(char *key);
-
 #endif /* __CR_OPTIONS_H__ */
