@@ -30,6 +30,7 @@
 #include "aio.h"
 #include "fault-injection.h"
 #include "syscall-codes.h"
+#include "signal.h"
 
 #include <string.h>
 #include <stdlib.h>
@@ -43,36 +44,28 @@
 #define MEMFD_FNAME	"CRIUMFD"
 #define MEMFD_FNAME_SZ	sizeof(MEMFD_FNAME)
 
-static int can_run_syscall(unsigned long ip, unsigned long start,
-			   unsigned long end, unsigned long pad)
-{
-	return ip >= start && ip < (end - BUILTIN_SYSCALL_SIZE - pad);
-}
-
-static int syscall_fits_vma_area(struct vma_area *vma_area, unsigned long pad)
-{
-	return can_run_syscall((unsigned long)vma_area->e->start,
-			       (unsigned long)vma_area->e->start,
-			       (unsigned long)vma_area->e->end,
-			       pad);
-}
-
-static struct vma_area *get_vma_by_ip(struct list_head *vma_area_list,
-				      unsigned long ip,
-				      unsigned long pad)
+unsigned long get_exec_start(struct vm_area_list *vmas)
 {
 	struct vma_area *vma_area;
 
-	list_for_each_entry(vma_area, vma_area_list, list) {
+	list_for_each_entry(vma_area, &vmas->h, list) {
+		unsigned long len;
+
 		if (vma_area->e->start >= kdat.task_size)
 			continue;
 		if (!(vma_area->e->prot & PROT_EXEC))
 			continue;
-		if (syscall_fits_vma_area(vma_area, pad))
-			return vma_area;
+
+		len = vma_area_len(vma_area);
+		if (len < BUILTIN_SYSCALL_SIZE + MEMFD_FNAME_SZ) {
+			pr_warn("Suspiciously short VMA @%#lx\n", (unsigned long)vma_area->e->start);
+			continue;
+		}
+
+		return vma_area->e->start;
 	}
 
-	return NULL;
+	return 0;
 }
 
 static inline int ptrace_get_regs(int pid, user_regs_struct_t *regs)
@@ -215,7 +208,7 @@ err:
 int __parasite_execute_syscall(struct parasite_ctl *ctl,
 		user_regs_struct_t *regs, const char *code_syscall)
 {
-	pid_t pid = ctl->pid.real;
+	pid_t pid = ctl->rpid;
 	int err;
 	u8 code_orig[BUILTIN_SYSCALL_SIZE];
 
@@ -236,7 +229,7 @@ int __parasite_execute_syscall(struct parasite_ctl *ctl,
 
 	if (ptrace_poke_area(pid, (void *)code_orig,
 			     (void *)ctl->syscall_ip, sizeof(code_orig))) {
-		pr_err("Can't restore syscall blob (pid: %d)\n", ctl->pid.real);
+		pr_err("Can't restore syscall blob (pid: %d)\n", ctl->rpid);
 		err = -1;
 	}
 
@@ -502,7 +495,7 @@ static int accept_tsock(struct parasite_ctl *ctl)
 static int parasite_init_daemon(struct parasite_ctl *ctl, struct ns_id *net)
 {
 	struct parasite_init_args *args;
-	pid_t pid = ctl->pid.real;
+	pid_t pid = ctl->rpid;
 	user_regs_struct_t regs;
 	struct ctl_msg m = { };
 
@@ -884,7 +877,7 @@ static bool task_in_parasite(struct parasite_ctl *ctl, user_regs_struct_t *regs)
 
 static int parasite_fini_seized(struct parasite_ctl *ctl)
 {
-	pid_t pid = ctl->pid.real;
+	pid_t pid = ctl->rpid;
 	user_regs_struct_t regs;
 	int status, ret = 0;
 	enum trace_flags flag;
@@ -1107,7 +1100,7 @@ int parasite_cure_local(struct parasite_ctl *ctl)
 
 	if (ctl->local_map) {
 		if (munmap(ctl->local_map, ctl->map_length)) {
-			pr_err("munmap failed (pid: %d)\n", ctl->pid.real);
+			pr_err("munmap failed (pid: %d)\n", ctl->rpid);
 			ret = -1;
 		}
 	}
@@ -1135,10 +1128,10 @@ int parasite_cure_seized(struct parasite_ctl *ctl)
 int parasite_unmap(struct parasite_ctl *ctl, unsigned long addr)
 {
 	user_regs_struct_t regs = ctl->orig.regs;
-	pid_t pid = ctl->pid.real;
+	pid_t pid = ctl->rpid;
 	int ret = -1;
 
-	ret = parasite_run(pid, PTRACE_SYSCALL, addr, NULL, &regs, &ctl->orig);
+	ret = parasite_run(pid, PTRACE_SYSCALL, addr, ctl->rstack, &regs, &ctl->orig);
 	if (ret)
 		goto err;
 
@@ -1151,10 +1144,9 @@ err:
 }
 
 /* If vma_area_list is NULL, a place for injecting syscall will not be set. */
-struct parasite_ctl *parasite_prep_ctl(pid_t pid, struct vm_area_list *vma_area_list)
+struct parasite_ctl *parasite_prep_ctl(pid_t pid, unsigned long exec_start)
 {
 	struct parasite_ctl *ctl = NULL;
-	struct vma_area *vma_area;
 
 	if (!arch_can_dump_task(pid))
 		goto err;
@@ -1173,22 +1165,9 @@ struct parasite_ctl *parasite_prep_ctl(pid_t pid, struct vm_area_list *vma_area_
 	if (get_thread_ctx(pid, &ctl->orig))
 		goto err;
 
-	ctl->pid.real	= pid;
-	ctl->pid.virt	= 0;
+	ctl->rpid = pid;
 
-	if (vma_area_list == NULL)
-		return ctl;
-
-	/* Search a place for injecting syscall */
-	vma_area = get_vma_by_ip(&vma_area_list->h, REG_IP(ctl->orig.regs),
-				 MEMFD_FNAME_SZ);
-	if (!vma_area) {
-		pr_err("No suitable VMA found to run parasite "
-		       "bootstrap code (pid: %d)\n", pid);
-		goto err;
-	}
-
-	ctl->syscall_ip	= vma_area->e->start;
+	ctl->syscall_ip = exec_start;
 	pr_debug("Parasite syscall_ip at %p\n", (void *)ctl->syscall_ip);
 
 	return ctl;
@@ -1206,13 +1185,13 @@ static int parasite_mmap_exchange(struct parasite_ctl *ctl, unsigned long size)
 				      PROT_READ | PROT_WRITE | PROT_EXEC,
 				      MAP_ANONYMOUS | MAP_SHARED, -1, 0);
 	if (!ctl->remote_map) {
-		pr_err("Can't allocate memory for parasite blob (pid: %d)\n", ctl->pid.real);
+		pr_err("Can't allocate memory for parasite blob (pid: %d)\n", ctl->rpid);
 		return -1;
 	}
 
 	ctl->map_length = round_up(size, page_size());
 
-	fd = open_proc_rw(ctl->pid.real, "map_files/%p-%p",
+	fd = open_proc_rw(ctl->rpid, "map_files/%p-%p",
 		 ctl->remote_map, ctl->remote_map + ctl->map_length);
 	if (fd < 0)
 		return -1;
@@ -1234,7 +1213,7 @@ static int parasite_memfd_exchange(struct parasite_ctl *ctl, unsigned long size)
 {
 	void *where = (void *)ctl->syscall_ip + BUILTIN_SYSCALL_SIZE;
 	u8 orig_code[MEMFD_FNAME_SZ] = MEMFD_FNAME;
-	pid_t pid = ctl->pid.real;
+	pid_t pid = ctl->rpid;
 	unsigned long sret = -ENOSYS;
 	int ret, fd, lfd;
 
@@ -1269,7 +1248,7 @@ static int parasite_memfd_exchange(struct parasite_ctl *ctl, unsigned long size)
 		return fd;
 
 	ctl->map_length = round_up(size, page_size());
-	lfd = open_proc_rw(ctl->pid.real, "fd/%d", fd);
+	lfd = open_proc_rw(ctl->rpid, "fd/%d", fd);
 	if (lfd < 0)
 		goto err_cure;
 
@@ -1344,7 +1323,7 @@ void parasite_ensure_args_size(unsigned long sz)
 
 static int parasite_start_daemon(struct parasite_ctl *ctl, struct pstree_item *item)
 {
-	pid_t pid = ctl->pid.real;
+	pid_t pid = ctl->rpid;
 
 	/*
 	 * Get task registers before going daemon, since the
@@ -1357,7 +1336,7 @@ static int parasite_start_daemon(struct parasite_ctl *ctl, struct pstree_item *i
 		return -1;
 	}
 
-	if (construct_sigframe(ctl->sigframe, ctl->rsigframe, item->core[0]))
+	if (construct_sigframe(ctl->sigframe, ctl->rsigframe, &ctl->orig.sigmask, item->core[0]))
 		return -1;
 
 	if (parasite_init_daemon(ctl, dmpi(item)->netns))
@@ -1375,7 +1354,13 @@ struct parasite_ctl *parasite_infect_seized(pid_t pid, struct pstree_item *item,
 
 	BUG_ON(item->threads[0].real != pid);
 
-	ctl = parasite_prep_ctl(pid, vma_area_list);
+	p = get_exec_start(vma_area_list);
+	if (!p) {
+		pr_err("No suitable VM found\n");
+		return NULL;
+	}
+
+	ctl = parasite_prep_ctl(pid, p);
 	if (!ctl)
 		return NULL;
 
