@@ -309,7 +309,7 @@ static int pie_atoi(char *str)
 	return ret;
 }
 
-static int get_proc_fd()
+static int get_proc_fd(void)
 {
 	int ret;
 	char buf[11];
@@ -346,7 +346,7 @@ static int get_proc_fd()
 	return open_detach_mount(proc_mountpoint);
 }
 
-static int parasite_get_proc_fd()
+static int parasite_get_proc_fd(void)
 {
 	int fd, ret;
 
@@ -605,7 +605,7 @@ static noinline void fini_sigreturn(unsigned long new_sp)
 	ARCH_RT_SIGRETURN(new_sp);
 }
 
-static int fini()
+static int fini(void)
 {
 	unsigned long new_sp;
 
@@ -614,7 +614,7 @@ static int fini()
 		mprotect_vmas(mprotect_args);
 	}
 
-	new_sp = (long)sigframe + SIGFRAME_OFFSET;
+	new_sp = (long)sigframe + RT_SIGFRAME_OFFSET(sigframe);
 	pr_debug("%ld: new_sp=%lx ip %lx\n", sys_gettid(),
 		  new_sp, RT_SIGFRAME_REGIP(sigframe));
 
