@@ -1,6 +1,7 @@
 #!/usr/bin/env python2
 # vim: noet
 import argparse
+import glob
 import os
 import subprocess
 import time
@@ -15,15 +16,17 @@ import linecache
 import random
 import string
 import imp
-import socket
 import fcntl
 import errno
 import datetime
 import yaml
+import criu as crpc
 
 os.chdir(os.path.dirname(os.path.abspath(__file__)))
 
 prev_line = None
+
+
 def traceit(f, e, a):
 	if e == "line":
 		lineno = f.f_lineno
@@ -44,28 +47,32 @@ def traceit(f, e, a):
 # sit in the same dir
 tests_root = None
 
+
 def clean_tests_root():
 	global tests_root
-	if tests_root:
-		os.rmdir(tests_root)
+	if tests_root and tests_root[0] == os.getpid():
+		os.rmdir(tests_root[1])
+
 
 def make_tests_root():
 	global tests_root
 	if not tests_root:
-		tests_root = tempfile.mkdtemp("", "criu-root-", "/tmp")
+		tests_root = (os.getpid(), tempfile.mkdtemp("", "criu-root-", "/tmp"))
 		atexit.register(clean_tests_root)
-	return tests_root
+	return tests_root[1]
 
 # Report generation
 
 report_dir = None
 
+
 def init_report(path):
 	global report_dir
 	report_dir = path
 	if not os.access(report_dir, os.F_OK):
 		os.makedirs(report_dir)
 
+
 def add_to_report(path, tgt_name):
 	global report_dir
 	if report_dir:
@@ -82,6 +89,7 @@ def add_to_report(path, tgt_name):
 				os.mkdir(os.path.dirname(tgt_path))
 			shutil.copy2(path, tgt_path)
 
+
 def add_to_output(path):
 	global report_dir
 	if not report_dir:
@@ -90,12 +98,31 @@ def add_to_output(path):
 	fdi = open(path, "r")
 	fdo = open(os.path.join(report_dir, "output"), "a")
 	while True:
-		buf = fdi.read(1<<20)
+		buf = fdi.read(1 << 20)
 		if not buf:
 			break
 		fdo.write(buf)
 
 
+prev_crash_reports = set(glob.glob("/tmp/zdtm-core-*.txt"))
+
+
+def check_core_files():
+	reports = set(glob.glob("/tmp/zdtm-core-*.txt")) - prev_crash_reports
+	if not reports:
+		return False
+
+	while subprocess.Popen("ps axf | grep 'abrt\.sh'", shell = True).wait() == 0:
+		time.sleep(1)
+
+	for i in reports:
+		add_to_report(i, os.path.basename(i))
+		print_sep(i)
+		print open(i).read()
+		print_sep(i)
+
+	return True
+
 # Arch we run on
 arch = os.uname()[4]
 
@@ -106,6 +133,7 @@ arch = os.uname()[4]
 #  uns -- user namespace, the same as above plus user namespace
 #
 
+
 class host_flavor:
 	def __init__(self, opts):
 		self.name = "host"
@@ -122,6 +150,7 @@ class host_flavor:
 	def clean():
 		pass
 
+
 class ns_flavor:
 	__root_dirs = ["/bin", "/sbin", "/etc", "/lib", "/lib64", "/dev", "/dev/pts", "/dev/net", "/tmp", "/usr", "/proc"]
 
@@ -151,10 +180,10 @@ class ns_flavor:
 
 		# This Mayakovsky-style code gets list of libraries a binary
 		# needs minus vdso and gate .so-s
-		libs = map(lambda x: x[1] == '=>' and x[2] or x[0],		\
-				map(lambda x: x.split(),			\
-					filter(lambda x: not xl.match(x),	\
-						map(lambda x: x.strip(),	\
+		libs = map(lambda x: x[1] == '=>' and x[2] or x[0],
+				map(lambda x: x.split(),
+					filter(lambda x: not xl.match(x),
+						map(lambda x: x.strip(),
 							filter(lambda x: x.startswith('\t'), ldd.stdout.readlines())))))
 		ldd.wait()
 
@@ -181,15 +210,16 @@ class ns_flavor:
 			os.mkdir(self.root + dir)
 			os.chmod(self.root + dir, 0777)
 
-		for ldir in [ "/bin", "/sbin", "/lib", "/lib64" ]:
+		for ldir in ["/bin", "/sbin", "/lib", "/lib64"]:
 			os.symlink(".." + ldir, self.root + "/usr" + ldir)
 
 		self.__mknod("tty", os.makedev(5, 0))
 		self.__mknod("null", os.makedev(1, 3))
 		self.__mknod("net/tun")
 		self.__mknod("rtc")
+		self.__mknod("autofs", os.makedev(10, 235))
 
-        def __copy_deps(self, deps):
+	def __copy_deps(self, deps):
 		for d in deps.split('|'):
 			if os.access(d, os.F_OK):
 				self.__copy_one(d)
@@ -216,8 +246,7 @@ class ns_flavor:
 
 	def fini(self):
 		if self.root_mounted:
-			subprocess.check_call(["mount", "--make-private", self.root])
-			subprocess.check_call(["umount", "-l", self.root])
+			subprocess.check_call(["./umount2", self.root])
 			self.root_mounted = False
 
 	@staticmethod
@@ -248,43 +277,53 @@ class userns_flavor(ns_flavor):
 		pass
 
 
-flavors = { 'h': host_flavor, 'ns': ns_flavor, 'uns': userns_flavor }
+flavors = {'h': host_flavor, 'ns': ns_flavor, 'uns': userns_flavor}
 
 #
 # Helpers
 #
 
+
 def encode_flav(f):
 	return (flavors.keys().index(f) + 128)
 
+
 def decode_flav(i):
-        i = i - 128
-        if flavors.has_key(i):
-                return flavors.keys()[i - 128]
-        return "unknown"
+	i = i - 128
+	if i in flavors:
+		return flavors.keys()[i - 128]
+	return "unknown"
+
 
 def tail(path):
 	p = subprocess.Popen(['tail', '-n1', path],
 			stdout = subprocess.PIPE)
-	return p.stdout.readline()
+	out = p.stdout.readline()
+	p.wait()
+	return out
+
 
 def rpidfile(path):
 	return open(path).readline().strip()
 
+
 def wait_pid_die(pid, who, tmo = 30):
 	stime = 0.1
 	while stime < tmo:
 		try:
 			os.kill(int(pid), 0)
-		except: # Died
+		except:  # Died
 			break
 
-		print "Wait for %s to die for %f" % (who, stime)
+		print "Wait for %s(%d) to die for %f" % (who, pid, stime)
 		time.sleep(stime)
 		stime *= 2
 	else:
+		subprocess.Popen(["ps", "-p", str(pid)]).wait()
+		subprocess.Popen(["ps", "axf", str(pid)]).wait()
 		raise test_fail_exc("%s die" % who)
 
+
 def test_flag(tdesc, flag):
 	return flag in tdesc.get('flags', '').split()
 
@@ -294,10 +333,12 @@ def test_flag(tdesc, flag):
 # test checks fail
 #
 
+
 class test_fail_exc:
 	def __init__(self, step):
 		self.step = step
 
+
 class test_fail_expected_exc:
 	def __init__(self, cr_action):
 		self.cr_action = cr_action
@@ -306,6 +347,7 @@ class test_fail_expected_exc:
 # A test from zdtm/ directory.
 #
 
+
 class zdtm_test:
 	def __init__(self, name, desc, flavor, freezer):
 		self.__name = name
@@ -315,16 +357,16 @@ class zdtm_test:
 		self.__pid = 0
 		self.__flavor = flavor
 		self.__freezer = freezer
-		self._bins = [ name ]
+		self._bins = [name]
 		self._env = {}
 		self._deps = desc.get('deps', [])
 		self.auto_reap = True
 
 	def __make_action(self, act, env = None, root = None):
-		sys.stdout.flush() # Not to let make's messages appear before ours
+		sys.stdout.flush()  # Not to let make's messages appear before ours
 		tpath = self.__name + '.' + act
-		s_args = ['make', '--no-print-directory', \
-			 	'-C', os.path.dirname(tpath), \
+		s_args = ['make', '--no-print-directory',
+				'-C', os.path.dirname(tpath),
 				      os.path.basename(tpath)]
 
 		if env:
@@ -359,7 +401,7 @@ class zdtm_test:
 			env['ZDTM_THREAD_BOMB'] = "5"
 
 		if not test_flag(self.__desc, 'suid'):
-			# Numbers should match those in criu_cli
+			# Numbers should match those in criu
 			env['ZDTM_UID'] = "18943"
 			env['ZDTM_GID'] = "58467"
 			env['ZDTM_GROUPS'] = "27495 48244"
@@ -379,7 +421,7 @@ class zdtm_test:
 				criu_dir = os.path.dirname(os.getcwd())
 				criu_dir_r = "%s%s" % (self.__flavor.root, criu_dir)
 
-				env['ZDTM_CRIU'] = os.path.dirname(os.getcwd());
+				env['ZDTM_CRIU'] = os.path.dirname(os.getcwd())
 				subprocess.check_call(["mkdir", "-p", criu_dir_r])
 
 		self.__make_action('pid', env, self.__flavor.root)
@@ -397,6 +439,7 @@ class zdtm_test:
 	def kill(self, sig = signal.SIGKILL):
 		self.__freezer.thaw()
 		if self.__pid:
+			print "Send the %d signal to  %s" % (sig, self.__pid)
 			os.kill(int(self.__pid), sig)
 			self.gone(sig == signal.SIGKILL)
 
@@ -404,11 +447,15 @@ class zdtm_test:
 
 	def stop(self):
 		self.__freezer.thaw()
-		self.getpid() # Read the pid from pidfile back
+		self.getpid()  # Read the pid from pidfile back
 		self.kill(signal.SIGTERM)
 
 		res = tail(self.__name + '.out')
-		if not 'PASS' in res.split():
+		if 'PASS' not in res.split():
+			if os.access(self.__name + '.out.inprogress', os.F_OK):
+				print_sep(self.__name + '.out.inprogress')
+				print open(self.__name + '.out.inprogress').read()
+				print_sep(self.__name + '.out.inprogress')
 			raise test_fail_exc("result check")
 
 	def getpid(self):
@@ -462,6 +509,8 @@ class zdtm_test:
 
 	@staticmethod
 	def available():
+		if not os.access("umount2", os.X_OK):
+			subprocess.check_call(["make", "umount2"])
 		if not os.access("zdtm_ct", os.X_OK):
 			subprocess.check_call(["make", "zdtm_ct"])
 		if not os.access("zdtm/lib/libzdtmtst.a", os.F_OK):
@@ -495,7 +544,7 @@ class inhfd_test:
 		if self.__peer_pid == 0:
 			os.setsid()
 
-			getattr(self.__fdtyp, "child_prep", lambda fd : None)(peer_file)
+			getattr(self.__fdtyp, "child_prep", lambda fd: None)(peer_file)
 
 			os.close(0)
 			os.close(1)
@@ -574,15 +623,15 @@ class groups_test(zdtm_test):
 
 		self._bins += self.__subs
 		self._deps += get_test_desc('zdtm/lib/groups')['deps']
-		self._env = { 'ZDTM_TESTS': self.__real_name }
+		self._env = {'ZDTM_TESTS': self.__real_name}
 
 	def __get_start_cmd(self, name):
 		tdir = os.path.dirname(name)
 		tname = os.path.basename(name)
 
 		s_args = ['make', '--no-print-directory', '-C', tdir]
-		subprocess.check_call(s_args + [ tname + '.cleanout' ])
-		s = subprocess.Popen(s_args + [ '--dry-run', tname + '.pid' ], stdout = subprocess.PIPE)
+		subprocess.check_call(s_args + [tname + '.cleanout'])
+		s = subprocess.Popen(s_args + ['--dry-run', tname + '.pid'], stdout = subprocess.PIPE)
 		cmd = s.stdout.readlines().pop().strip()
 		s.wait()
 
@@ -606,11 +655,11 @@ class groups_test(zdtm_test):
 
 		for test in self.__subs:
 			res = tail(test + '.out')
-			if not 'PASS' in res.split():
+			if 'PASS' not in res.split():
 				raise test_fail_exc("sub %s result check" % test)
 
 
-test_classes = { 'zdtm': zdtm_test, 'inhfd': inhfd_test, 'groups': groups_test }
+test_classes = {'zdtm': zdtm_test, 'inhfd': inhfd_test, 'groups': groups_test}
 
 #
 # CRIU when launched using CLI
@@ -618,7 +667,87 @@ test_classes = { 'zdtm': zdtm_test, 'inhfd': inhfd_test, 'groups': groups_test }
 
 criu_bin = "../criu/criu"
 join_ns_file = '/run/netns/zdtm_netns'
+
+
 class criu_cli:
+	@staticmethod
+	def run(action, args, fault = None, strace = [], preexec = None):
+		env = None
+		if fault:
+			print "Forcing %s fault" % fault
+			env = dict(os.environ, CRIU_FAULT = fault)
+		cr = subprocess.Popen(strace + [criu_bin, action] + args, env = env, preexec_fn = preexec)
+		return cr.wait()
+
+
+class criu_rpc:
+	@staticmethod
+	def __set_opts(criu, args, ctx):
+		while len(args) != 0:
+			arg = args.pop(0)
+			if arg == '-v4':
+				criu.opts.log_level = 4
+				continue
+			if arg == '-o':
+				criu.opts.log_file = args.pop(0)
+				continue
+			if arg == '-D':
+				criu.opts.images_dir_fd = os.open(args.pop(0), os.O_DIRECTORY)
+				ctx['imgd'] = criu.opts.images_dir_fd
+				continue
+			if arg == '-t':
+				criu.opts.pid = int(args.pop(0))
+				continue
+			if arg == '--pidfile':
+				ctx['pidf'] = args.pop(0)
+				continue
+			if arg == '--timeout':
+				criu.opts.timeout = int(args.pop(0))
+				continue
+			if arg == '--restore-detached':
+				# Set by service by default
+				ctx['rd'] = True
+				continue
+			if arg == '--root':
+				criu.opts.root = args.pop(0)
+				continue
+
+			raise test_fail_exc('RPC for %s required' % arg)
+
+	@staticmethod
+	def run(action, args, fault = None, strace = [], preexec = None):
+		if fault:
+			raise test_fail_exc('RPC and FAULT not supported')
+		if strace:
+			raise test_fail_exc('RPC and SAT not supported')
+		if preexec:
+			raise test_fail_exc('RPC and PREEXEC not supported')
+
+		ctx = {}  # Object used to keep info untill action is done
+		criu = crpc.criu()
+		criu.use_binary(criu_bin)
+		criu_rpc.__set_opts(criu, args, ctx)
+
+		if action == 'dump':
+			criu.dump()
+		elif action == 'restore':
+			if 'rd' not in ctx:
+				raise test_fail_exc('RPC Non-detached restore is impossible')
+
+			res = criu.restore()
+			pidf = ctx.get('pidf')
+			if pidf:
+				open(pidf, 'w').write('%d\n' % res.pid)
+		else:
+			raise test_fail_exc('RPC for %s required' % action)
+
+		imgd = ctx.get('imgd')
+		if imgd:
+			os.close(imgd)
+		return 0
+
+
+class criu:
 	def __init__(self, opts):
 		self.__test = None
 		self.__dump_path = None
@@ -631,7 +760,10 @@ class criu_cli:
 		self.__script = opts['script']
 		self.__sat = (opts['sat'] and True or False)
 		self.__dedup = (opts['dedup'] and True or False)
+		self.__mdedup = (opts['noauto_dedup'] and True or False)
 		self.__user = (opts['user'] and True or False)
+		self.__leave_stopped = (opts['stop'] and True or False)
+		self.__criu = (opts['rpc'] and criu_rpc or criu_cli)
 
 	def logs(self):
 		return self.__dump_path
@@ -658,15 +790,6 @@ class criu_cli:
 	def __ddir(self):
 		return os.path.join(self.__dump_path, "%d" % self.__iter)
 
-	@staticmethod
-	def __criu(action, args, fault = None, strace = [], preexec = None):
-		env = None
-		if fault:
-			print "Forcing %s fault" % fault
-			env = dict(os.environ, CRIU_FAULT = fault)
-		cr = subprocess.Popen(strace + [criu_bin, action] + args, env = env, preexec_fn = preexec)
-		return cr.wait()
-
 	def set_user_id(self):
 		# Numbers should match those in zdtm_test
 		os.setresgid(58467, 58467, 58467)
@@ -688,17 +811,20 @@ class criu_cli:
 			print_fname(fname, 'strace')
 			strace = ["strace", "-o", fname, '-T']
 			if action == 'restore':
-				strace += [ '-f' ]
-				s_args += [ '--action-script', os.getcwd() + '/../scripts/fake-restore.sh' ]
+				strace += ['-f']
+				s_args += ['--action-script', os.getcwd() + '/../scripts/fake-restore.sh']
 
 		if self.__script:
 			s_args += ['--action-script', self.__script]
 
-		preexec = self.__user and self.set_user_id or None
+		if action == "restore":
+			preexec = None
+		else:
+			preexec = self.__user and self.set_user_id or None
 
 		__ddir = self.__ddir()
 
-		ret = self.__criu(action, s_args, self.__fault, strace, preexec)
+		ret = self.__criu.run(action, s_args, self.__fault, strace, preexec)
 		grep_errors(os.path.join(__ddir, log))
 		if ret != 0:
 			if self.__fault and int(self.__fault) < 128:
@@ -712,9 +838,9 @@ class criu_cli:
 					# on restore we move only a log file, because we need images
 					os.rename(os.path.join(__ddir, log), os.path.join(__ddir, log + ".fail"))
 				# try again without faults
-		                print "Run criu " + action
-				ret = self.__criu(action, s_args, False, strace, preexec)
-		                grep_errors(os.path.join(__ddir, log))
+				print "Run criu " + action
+				ret = self.__criu.run(action, s_args, False, strace, preexec)
+				grep_errors(os.path.join(__ddir, log))
 				if ret == 0:
 					return
 			if self.__test.blocking() or (self.__sat and action == 'restore'):
@@ -735,9 +861,9 @@ class criu_cli:
 		if self.__page_server:
 			print "Adding page server"
 
-			ps_opts = [ "--port", "12345", "--daemon", "--pidfile", "ps.pid" ]
+			ps_opts = ["--port", "12345", "--daemon", "--pidfile", "ps.pid"]
 			if self.__dedup:
-				ps_opts += [ "--auto-dedup" ]
+				ps_opts += ["--auto-dedup"]
 
 			self.__criu_act("page-server", opts = ps_opts)
 			a_opts += ["--page-server", "--address", "127.0.0.1", "--port", "12345"]
@@ -745,16 +871,25 @@ class criu_cli:
 		a_opts += self.__test.getdopts()
 
 		if self.__dedup:
-			a_opts += [ "--auto-dedup" ]
+			a_opts += ["--auto-dedup"]
 
-		a_opts += [ "--timeout", "10" ]
+		a_opts += ["--timeout", "10"]
 
 		criu_dir = os.path.dirname(os.getcwd())
 		if os.getenv("GCOV"):
 			a_opts.append("--ext-mount-map")
 			a_opts.append("%s:zdtm" % criu_dir)
 
+		if self.__leave_stopped:
+			a_opts += ['--leave-stopped']
+
 		self.__criu_act(action, opts = a_opts + opts)
+		if self.__mdedup and self.__iter > 1:
+			self.__criu_act("dedup", opts = [])
+
+		if self.__leave_stopped:
+			pstree_check_stopped(self.__test.getpid())
+			pstree_signal(self.__test.getpid(), signal.SIGKILL)
 
 		if self.__page_server:
 			wait_pid_die(int(rpidfile(self.__ddir() + "/ps.pid")), "page server")
@@ -774,11 +909,19 @@ class criu_cli:
 		if os.getenv("GCOV"):
 			r_opts.append("--ext-mount-map")
 			r_opts.append("zdtm:%s" % criu_dir)
+
+		if self.__leave_stopped:
+			r_opts += ['--leave-stopped']
+
 		self.__criu_act("restore", opts = r_opts + ["--restore-detached"])
 
+		if self.__leave_stopped:
+			pstree_check_stopped(self.__test.getpid())
+			pstree_signal(self.__test.getpid(), signal.SIGCONT)
+
 	@staticmethod
 	def check(feature):
-		return criu_cli.__criu("check", ["-v0", "--feature", feature]) == 0
+		return criu_cli.run("check", ["-v0", "--feature", feature]) == 0
 
 	@staticmethod
 	def available():
@@ -801,6 +944,7 @@ def try_run_hook(test, args):
 
 do_sbs = False
 
+
 def init_sbs():
 	if sys.stdout.isatty():
 		global do_sbs
@@ -808,18 +952,20 @@ def init_sbs():
 	else:
 		print "Can't do step-by-step in this runtime"
 
+
 def sbs(what):
 	if do_sbs:
 		raw_input("Pause at %s. Press any key to continue." % what)
 
+
 #
 # Main testing entity -- dump (probably with pre-dumps) and restore
 #
-
 def iter_parm(opt, dflt):
 	x = ((opt or str(dflt)) + ":0").split(':')
 	return (xrange(0, int(x[0])), float(x[1]))
 
+
 def cr(cr_api, test, opts):
 	if opts['nocr']:
 		return
@@ -855,12 +1001,12 @@ def cr(cr_api, test, opts):
 # Additional checks that can be done outside of test process
 
 def get_visible_state(test):
-	maps	= {}
-	files	= {}
-	mounts	= {}
+	maps = {}
+	files = {}
+	mounts = {}
 
-	if not getattr(test, "static", lambda : False)() or \
-	   not getattr(test, "ns", lambda : False)():
+	if not getattr(test, "static", lambda: False)() or \
+	   not getattr(test, "ns", lambda: False)():
 		return ({}, {}, {})
 
 	r = re.compile('^[0-9]+$')
@@ -868,16 +1014,25 @@ def get_visible_state(test):
 	for pid in pids:
 		files[pid] = set(os.listdir("/proc/%s/root/proc/%s/fd" % (test.getpid(), pid)))
 
-		cmaps = [[0, 0]]
+		cmaps = [[0, 0, ""]]
 		last = 0
 		for mp in open("/proc/%s/root/proc/%s/maps" % (test.getpid(), pid)):
 			m = map(lambda x: int('0x' + x, 0), mp.split()[0].split('-'))
-			if cmaps[last][1] == m[0]:
+
+			m.append(mp.split()[1])
+
+			f = "/proc/%s/root/proc/%s/map_files/%s" % (test.getpid(), pid, mp.split()[0])
+			if os.access(f, os.F_OK):
+				st = os.lstat(f)
+				m.append(oct(st.st_mode))
+
+			if cmaps[last][1] == m[0] and cmaps[last][2] == m[2]:
 				cmaps[last][1] = m[1]
 			else:
 				cmaps.append(m)
 				last += 1
-		maps[pid] = set(map(lambda x: '%x-%x' % (x[0], x[1]), cmaps))
+
+		maps[pid] = set(map(lambda x: '%x-%x %s' % (x[0], x[1], x[2:]), cmaps))
 
 		cmounts = []
 		try:
@@ -890,6 +1045,7 @@ def get_visible_state(test):
 		mounts[pid] = cmounts
 	return files, maps, mounts
 
+
 def check_visible_state(test, state, opts):
 	new = get_visible_state(test)
 
@@ -906,7 +1062,7 @@ def check_visible_state(test, state, opts):
 		if old_maps != new_maps:
 			print "%s: Old maps lost: %s" % (pid, old_maps - new_maps)
 			print "%s: New maps appeared: %s" % (pid, new_maps - old_maps)
-			if not opts['fault']: # skip parasite blob
+			if not opts['fault']:  # skip parasite blob
 				raise test_fail_exc("maps compare")
 
 		old_mounts = state[2][pid]
@@ -922,6 +1078,13 @@ def check_visible_state(test, state, opts):
 			print "%s: New mounts appeared: %s" % (pid, new_mounts)
 			raise test_fail_exc("mounts compare")
 
+	if '--link-remap' in test.getdopts():
+		import glob
+		link_remap_list = glob.glob(os.path.dirname(test.getname()) + '/link_remap*')
+		if link_remap_list:
+			print "%s: link-remap files left: %s" % (test.getname(), link_remap_list)
+			raise test_fail_exc("link remaps left")
+
 
 class noop_freezer:
 	def __init__(self):
@@ -968,10 +1131,10 @@ class cg_freezer:
 			self.__set_state('THAWED')
 
 	def getdopts(self):
-		return [ '--freeze-cgroup', self.__path, '--manage-cgroups' ]
+		return ['--freeze-cgroup', self.__path, '--manage-cgroups']
 
 	def getropts(self):
-		return [ '--manage-cgroups' ]
+		return ['--manage-cgroups']
 
 
 def get_freezer(desc):
@@ -982,6 +1145,7 @@ def get_freezer(desc):
 	fr = cg_freezer(path = fd[0], state = fd[1])
 	return fr
 
+
 def cmp_ns(ns1, match, ns2, msg):
 	ns1_ino = os.stat(ns1).st_ino
 	ns2_ino = os.stat(ns2).st_ino
@@ -994,6 +1158,69 @@ def check_joinns_state(t):
 	cmp_ns("/proc/%s/ns/net" % t.getpid(), "!=", join_ns_file, "join-ns")
 
 
+def pstree_each_pid(root_pid):
+	f_children_path = "/proc/{0}/task/{0}/children".format(root_pid)
+	child_pids = []
+	try:
+		with open(f_children_path, "r") as f_children:
+			pid_line = f_children.readline().strip(" \n")
+			if pid_line:
+				child_pids += pid_line.split(" ")
+	except:
+		return  # process is dead
+
+	yield root_pid
+	for child_pid in child_pids:
+		for pid in pstree_each_pid(child_pid):
+			yield pid
+
+
+def is_proc_stopped(pid):
+	def get_thread_status(thread_dir):
+		try:
+			with open(os.path.join(thread_dir, "status")) as f_status:
+				for line in f_status.readlines():
+					if line.startswith("State:"):
+						return line.split(":", 1)[1].strip().split(" ")[0]
+		except:
+			pass  # process is dead
+		return None
+
+	def is_thread_stopped(status):
+		return (status is None) or (status == "T") or (status == "Z")
+
+	tasks_dir = "/proc/%s/task" % pid
+	thread_dirs = []
+	try:
+		thread_dirs = os.listdir(tasks_dir)
+	except:
+		pass  # process is dead
+
+	for thread_dir in thread_dirs:
+		thread_status = get_thread_status(os.path.join(tasks_dir, thread_dir))
+		if not is_thread_stopped(thread_status):
+			return False
+
+	if not is_thread_stopped(get_thread_status("/proc/%s" % pid)):
+		return False
+
+	return True
+
+
+def pstree_check_stopped(root_pid):
+	for pid in pstree_each_pid(root_pid):
+		if not is_proc_stopped(pid):
+			raise test_fail_exc("CRIU --leave-stopped %s" % pid)
+
+
+def pstree_signal(root_pid, signal):
+	for pid in pstree_each_pid(root_pid):
+		try:
+			os.kill(int(pid), signal)
+		except:
+			pass  # process is dead
+
+
 def do_run_test(tname, tdesc, flavs, opts):
 	tcname = tname.split('/')[0]
 	tclass = test_classes.get(tcname, None)
@@ -1015,7 +1242,7 @@ def do_run_test(tname, tdesc, flavs, opts):
 			continue
 		flav = flavors[f](opts)
 		t = tclass(tname, tdesc, flav, fcg)
-		cr_api = criu_cli(opts)
+		cr_api = criu(opts)
 
 		try:
 			t.start()
@@ -1048,6 +1275,7 @@ def do_run_test(tname, tdesc, flavs, opts):
 				cr_api.cleanup()
 			print_sep("Test %s PASS" % tname)
 
+
 class launcher:
 	def __init__(self, opts, nr_tests):
 		self.__opts = opts
@@ -1103,10 +1331,10 @@ class launcher:
 		self.__nr += 1
 		self.__show_progress()
 
-		nd = ('nocr', 'norst', 'pre', 'iters', 'page_server', 'sibling', \
-				'fault', 'keep_img', 'report', 'snaps', 'sat', 'script', \
-				'join_ns', 'dedup', 'sbs', 'freezecg', 'user', 'dry_run')
-		arg = repr((name, desc, flavor, { d: self.__opts[d] for d in nd }))
+		nd = ('nocr', 'norst', 'pre', 'iters', 'page_server', 'sibling', 'stop',
+				'fault', 'keep_img', 'report', 'snaps', 'sat', 'script', 'rpc',
+				'join_ns', 'dedup', 'sbs', 'freezecg', 'user', 'dry_run', 'noauto_dedup')
+		arg = repr((name, desc, flavor, {d: self.__opts[d] for d in nd}))
 
 		if self.__use_log:
 			logf = name.replace('/', '_') + ".log"
@@ -1115,10 +1343,10 @@ class launcher:
 			logf = None
 			log = None
 
-		sub = subprocess.Popen(["./zdtm_ct", "zdtm.py"], \
-				env = dict(os.environ, CR_CT_TEST_INFO = arg ), \
+		sub = subprocess.Popen(["./zdtm_ct", "zdtm.py"],
+				env = dict(os.environ, CR_CT_TEST_INFO = arg),
 				stdout = log, stderr = subprocess.STDOUT, close_fds = True)
-		self.__subs[sub.pid] = { 'sub': sub, 'log': logf, 'name': name }
+		self.__subs[sub.pid] = {'sub': sub, 'log': logf, 'name': name}
 
 		if test_flag(desc, 'excl'):
 			self.wait()
@@ -1133,7 +1361,7 @@ class launcher:
 				failed_flavor = decode_flav(os.WEXITSTATUS(status))
 				if self.__file_report:
 					testline = "not ok %d - %s # flavor %s" % (self.__runtest, sub['name'], failed_flavor)
-					details = { 'output': open(sub['log']).read() }
+					details = {'output': open(sub['log']).read()}
 					print >> self.__file_report, testline
 					print >> self.__file_report, yaml.dump(details, explicit_start=True, explicit_end=True, default_style='|')
 				if sub['log']:
@@ -1170,29 +1398,32 @@ class launcher:
 
 	def finish(self):
 		self.__wait_all()
+		if not opts['fault'] and check_core_files():
+			self.__fail = True
 		if self.__file_report:
 			self.__file_report.close()
 		if self.__fail:
 			print_sep("FAIL", "#")
 			sys.exit(1)
 
+
 def all_tests(opts):
 	desc = eval(open(opts['set'] + '.desc').read())
-	lst = subprocess.Popen(['find', desc['dir'], '-type', 'f', '-executable' ], \
+	lst = subprocess.Popen(['find', desc['dir'], '-type', 'f', '-executable'],
 			stdout = subprocess.PIPE)
 	excl = map(lambda x: os.path.join(desc['dir'], x), desc['exclude'])
-	tlist = filter(lambda x: \
-			not x.endswith('.checkskip') and \
-			not x.endswith('.hook') and \
-			not x in excl, \
-				map(lambda x: x.strip(), lst.stdout.readlines()) \
-		)
+	tlist = filter(lambda x:
+			not x.endswith('.checkskip') and
+			not x.endswith('.hook') and
+			x not in excl,
+			map(lambda x: x.strip(), lst.stdout.readlines())
+			)
 	lst.wait()
 	return tlist
 
 
 # Descriptor for abstract test not in list
-default_test={ }
+default_test = {}
 
 
 def get_test_desc(tname):
@@ -1204,13 +1435,14 @@ def get_test_desc(tname):
 
 
 def self_checkskip(tname):
-	chs = tname  + '.checkskip'
+	chs = tname + '.checkskip'
 	if os.access(chs, os.X_OK):
 		ch = subprocess.Popen([chs])
 		return not ch.wait() == 0
 
 	return False
 
+
 def print_fname(fname, typ):
 	print "=[%s]=> %s" % (typ, fname)
 
@@ -1218,6 +1450,7 @@ def print_fname(fname, typ):
 def print_sep(title, sep = "=", width = 80):
 	print (" " + title + " ").center(width, sep)
 
+
 def grep_errors(fname):
 	first = True
 	for l in open(fname):
@@ -1230,14 +1463,15 @@ def grep_errors(fname):
 	if not first:
 		print_sep("ERROR OVER", "-", 60)
 
+
 def run_tests(opts):
 	excl = None
 	features = {}
 
 	if opts['pre'] or opts['snaps']:
-		if not criu_cli.check("mem_dirty_track"):
+		if not criu.check("mem_dirty_track"):
 			print "Tracking memory is not available"
-			return;
+			return
 
 	if opts['keep_going'] and (not opts['all']):
 		print "[WARNING] Option --keep-going is more useful with option --all."
@@ -1279,9 +1513,9 @@ def run_tests(opts):
 
 	if opts['join_ns']:
 		if subprocess.Popen(["ip", "netns", "add", "zdtm_netns"]).wait():
-                    raise Exception("Unable to create a network namespace")
+			raise Exception("Unable to create a network namespace")
 		if subprocess.Popen(["ip", "netns", "exec", "zdtm_netns", "ip", "link", "set", "up", "dev", "lo"]).wait():
-                    raise Exception("ip link set up dev lo")
+			raise Exception("ip link set up dev lo")
 
 	l = launcher(opts, len(torun))
 	try:
@@ -1303,9 +1537,9 @@ def run_tests(opts):
 
 			feat = tdesc.get('feature', None)
 			if feat:
-				if not features.has_key(feat):
+				if feat not in features:
 					print "Checking feature %s" % feat
-					features[feat] = criu_cli.check(feat)
+					features[feat] = criu.check(feat)
 
 				if not features[feat]:
 					l.skip(t, "no %s feature" % feat)
@@ -1334,13 +1568,13 @@ def run_tests(opts):
 				run_flavs = set(test_flavs) & set(opts_flavs)
 			else:
 				run_flavs = set([test_flavs.pop()])
-			if not criu_cli.check("userns"):
+			if not criu.check("userns"):
 				run_flavs -= set(['uns'])
 			if opts['user']:
 				# FIXME -- probably uns will make sense
 				run_flavs -= set(['ns', 'uns'])
 
-			#remove ns and uns flavor in join_ns
+			# remove ns and uns flavor in join_ns
 			if opts['join_ns']:
 				run_flavs -= set(['ns', 'uns'])
 
@@ -1355,6 +1589,7 @@ def run_tests(opts):
 
 sti_fmt = "%-40s%-10s%s"
 
+
 def show_test_info(t):
 	tdesc = get_test_desc(t)
 	flavs = tdesc.get('flavor', '')
@@ -1371,7 +1606,7 @@ def list_tests(opts):
 
 class group:
 	def __init__(self, tname, tdesc):
-		self.__tests = [ tname ]
+		self.__tests = [tname]
 		self.__desc = tdesc
 		self.__deps = set()
 
@@ -1404,8 +1639,8 @@ class group:
 	# which will call all tests' scripts in turn
 	def __dump_meta(self, fname, ext):
 		scripts = filter(lambda names: os.access(names[1], os.X_OK),
-		                 map(lambda test: (test, test + ext),
-		                     self.__tests))
+				map(lambda test: (test, test + ext),
+				self.__tests))
 		if scripts:
 			f = open(fname + ext, "w")
 			f.write("#!/bin/sh -e\n")
@@ -1436,6 +1671,7 @@ class group:
 		self.__dump_meta(fname, '.checkskip')
 		self.__dump_meta(fname, '.hook')
 
+
 def group_tests(opts):
 	excl = None
 	groups = []
@@ -1473,7 +1709,7 @@ def group_tests(opts):
 	suf = opts['name'] or 'group'
 
 	for g in groups:
-		if maxs > 1 and g.size() == 1: # Not much point in group test for this
+		if maxs > 1 and g.size() == 1:  # Not much point in group test for this
 			continue
 
 		fn = os.path.join("groups", "%s.%d" % (suf, nr))
@@ -1495,7 +1731,7 @@ def clean_stuff(opts):
 # main() starts here
 #
 
-if os.environ.has_key('CR_CT_TEST_INFO'):
+if 'CR_CT_TEST_INFO' in os.environ:
 	# Fork here, since we're new pidns init and are supposed to
 	# collect this namespace's zombies
 	status = 0
@@ -1511,7 +1747,7 @@ if os.environ.has_key('CR_CT_TEST_INFO'):
 					status = os.WEXITSTATUS(status)
 				else:
 					status = 1
-				break;
+				break
 
 	sys.exit(status)
 
@@ -1535,21 +1771,24 @@ rp.add_argument("--join-ns", help = "Restore tests and join existing namespace",
 rp.add_argument("--pre", help = "Do some pre-dumps before dump (n[:pause])")
 rp.add_argument("--snaps", help = "Instead of pre-dumps do full dumps", action = 'store_true')
 rp.add_argument("--dedup", help = "Auto-deduplicate images on iterations", action = 'store_true')
+rp.add_argument("--noauto-dedup", help = "Manual deduplicate images on iterations", action = 'store_true')
 rp.add_argument("--nocr", help = "Do not CR anything, just check test works", action = 'store_true')
 rp.add_argument("--norst", help = "Don't restore tasks, leave them running after dump", action = 'store_true')
+rp.add_argument("--stop", help = "Check that --leave-stopped option stops ps tree.", action = 'store_true')
 rp.add_argument("--iters", help = "Do CR cycle several times before check (n[:pause])")
 rp.add_argument("--fault", help = "Test fault injection")
 rp.add_argument("--sat", help = "Generate criu strace-s for sat tool (restore is fake, images are kept)", action = 'store_true')
 rp.add_argument("--sbs", help = "Do step-by-step execution, asking user for keypress to continue", action = 'store_true')
 rp.add_argument("--freezecg", help = "Use freeze cgroup (path:state)")
 rp.add_argument("--user", help = "Run CRIU as regular user", action = 'store_true')
+rp.add_argument("--rpc", help = "Run CRIU via RPC rather than CLI", action = 'store_true')
 
 rp.add_argument("--page-server", help = "Use page server dump", action = 'store_true')
 rp.add_argument("-p", "--parallel", help = "Run test in parallel")
 rp.add_argument("--dry-run", help="Don't run tests, just pretend to", action='store_true')
 rp.add_argument("--script", help="Add script to get notified by criu")
 rp.add_argument("-k", "--keep-img", help = "Whether or not to keep images after test",
-		choices = [ 'always', 'never', 'failed' ], default = 'failed')
+		choices = ['always', 'never', 'failed'], default = 'failed')
 rp.add_argument("--report", help = "Generate summary report in directory")
 rp.add_argument("--keep-going", help = "Keep running tests in spite of failures", action = 'store_true')
 
@@ -1565,7 +1804,7 @@ gp.add_argument("-x", "--exclude", help = "Exclude tests from --all run", action
 
 cp = sp.add_parser("clean", help = "Clean something")
 cp.set_defaults(action = clean_stuff)
-cp.add_argument("what", choices = [ 'nsroot' ])
+cp.add_argument("what", choices = ['nsroot'])
 
 opts = vars(p.parse_args())
 if opts.get('sat', False):
@@ -1574,7 +1813,7 @@ if opts.get('sat', False):
 if opts['debug']:
 	sys.settrace(traceit)
 
-criu_cli.available()
+criu.available()
 for tst in test_classes.values():
 	tst.available()
 
